"""A collection of regression functions."""
from collections import namedtuple

import numpy as _np
from scipy import stats as _stats


def nanlinregress(x, y):
    """
    NaN-ignoring linear regression.

    Parameters
    ----------
    x, y : array_like
        Two sets of measurements. Both arrays should have the same length.

    Returns
    -------
    slope : float
        Slope of the regression line.
    intercept : float
        Intercept of the regression line.
    rvalue : float
        Pearson correlation coefficient.
    pvalue : float
        Two-sided p-value for a hypothesis test with the null hypothesis that
        the slope is zero.
    stderr : float
        Standard error of the estimated slope.

    """
    xfinite = x[_np.isfinite(x) & _np.isfinite(y)]
    yfinite = y[_np.isfinite(x) & _np.isfinite(y)]
    return _stats.linregress(xfinite, yfinite)


def linreg_zerointercept(x, y):
    """
    Do a linear regression with the intercept forced at zero.

    Parameters
    ----------
    x, y : array_like
        Two sets of measurements. Both arrays should have the same length.

    Returns
    -------
    slope : float
        Slope of the regression line.
    intercept : float
        Intercept of the regression line, forced to be zero.
    rvalue : float
        Pearson correlation coefficient.
    pvalue : float
        Two-sided p-value for a hypothesis test with the null hypothesis that
        the slope is zero.
    stderr : float
        Standard error of the estimated slope.

    """
    # extract finite values only
    xfinite = x[_np.isfinite(x) & _np.isfinite(y)]
    yfinite = y[_np.isfinite(x) & _np.isfinite(y)]

    n = xfinite.size
    df = n - 2  # degree of freedom
    intercept = 0.0  # intercept is forced to be zero by definition

    slope = _np.sum(yfinite * xfinite) / _np.sum(xfinite * xfinite)
    ypred = slope * xfinite  # predicted y
    rvalue, _ = _stats.pearsonr(yfinite, ypred)
    # standard error of the slope
    stderr = _np.sqrt(
        _np.sum((ypred - yfinite) ** 2.0)
        / (df * _np.sum((xfinite - _np.mean(xfinite)) ** 2.0))
    )
    pvalue = 2 * _stats.distributions.t.sf(_np.abs(slope / stderr), df)

    LinregZeroInterceptResult = namedtuple(
        "LinregressNoInterceptResult",
        ("slope", "intercept", "rvalue", "pvalue", "stderr"),
    )
    return LinregZeroInterceptResult(slope, intercept, rvalue, pvalue, stderr)



"""
==========================================
Leaf-scale processes (:mod:`ecoflux.leaf`)
==========================================

.. currentmodule:: ecoflux.leaf

Leaf-scale exchange of mass and energy.

"""
from . import light_response  # noqa



"""
============================================
Physical chemistry (:mod:`ecoflux.physchem`)
============================================

.. currentmodule:: ecoflux.physchem

Functions to calculate physicochemical properties.

"""
from . import sat_vap  # noqa



"""A collection of time series functions."""
from collections import namedtuple

import numpy as _np

from .summary import zscore


def running_std(series, window_size):
    """
    Calculate the running standard deviation.

    Parameters
    ----------
    series : array_like
        Input 1D time series.
    window_size : int
        Size of the moving window.

    Returns
    -------
    rnstd_series : array_like
        Time series of running standard deviation, same length to the input.

    Raises
    ------
    ValueError
        If the `window_size` is larger than the size of the input time series.

    See Also
    --------
    `running_zscore` : Running Z-score function.

    """
    n = series.size
    if n < window_size:
        raise ValueError("Window size larger than data size.")

    left_size = (int(window_size) - 1) // 2
    right_size = window_size - 1 - left_size
    rnstd_series = [
        _np.nanstd(
            series[extract_window(series, i, left_size, right_size)], ddof=1
        )
        for i in range(n)
    ]
    rnstd_series = _np.array(rnstd_series)
    return rnstd_series


def running_zscore(series, window_size, robust_zscore=False):
    """
    Calculate the running Z-score.

    Parameters
    ----------
    series : array_like
        Input 1D time series.
    window : int
        Size of the moving window.
    robust_zscore : bool, optional
        If `True`, use the Iglewicz-Hoaglin robust Z-score [IH93]_ instead of
        the original Z-score. Default is `False`.

    Returns
    -------
    zscore_series : array_like
        The calculated moving Z-score of the series.

    Raises
    ------
    ValueError
        If the `window_size` is larger than the size of the input time series.

    See Also
    --------
    `running_std` : Running standard deviation function.

    References
    ----------
    .. [IH93] Iglewicz, Boris and Hoaglin, David C. (1993). How to Detect and
        Handle Outliers. ASQC Quality Press, Milwaukee, WI, 1993.

    """
    n = series.size
    if n < window_size:
        raise ValueError("Window size larger than data size.")

    left_size = (int(window_size) - 1) // 2
    right_size = window_size - 1 - left_size
    zscore_series = [
        zscore(
            series[extract_window(series, i, left_size, right_size)],
            robust_zscore=robust_zscore,
        )[left_size]
        for i in range(n)
    ]
    zscore_series = _np.array(zscore_series)
    return zscore_series


def extract_window(series, i, left_size, right_size):
    """A helper function to return indices of the window centered at i."""
    n = series.size
    if i <= left_size:
        window_idx = [0] * (left_size - i) + list(range(0, right_size + i + 1))
    elif i + right_size + 1 > n:
        window_idx = list(range(i - left_size, n)) + [n - 1] * (
            right_size + i + 1 - n
        )
    else:
        window_idx = list(range(i - left_size, i + right_size + 1))
    return window_idx


def hourly_median(hours, series, all_hours=True):
    """
    Calculate hourly binned medians of a time series.

    Parameters
    ----------
    hours : array_like
        Time series of the hour number. Must be of the same length as `series`.
    series : array_like
        Time series of the data.
    all_hours : bool, optional
        Default is `True` to consider 24 hours. If `False`, only consider the
        hours that are present in the `hours` input.

    Returns
    -------
    hour_level : array_like
        Hour levels.
    median : array_like
        Median values by the hour.
    q1 : array_like
        First quartile values by the hour.
    q3 : array_like
        Third quartile values by the hour.

    See Also
    --------
    `hourly_avg` : Hourly binned average function.

    """
    if all_hours:
        hour_level = _np.arange(24)
    else:
        hour_level = _np.unique(hours)
    med_hr = _np.zeros(hour_level.size) + _np.nan
    q1_hr = _np.zeros(hour_level.size) + _np.nan
    q3_hr = _np.zeros(hour_level.size) + _np.nan
    for i in range(hour_level.size):
        med_hr[i] = _np.nanmedian(series[hours == hour_level[i]])
        q1_hr[i], q3_hr[i] = _np.nanpercentile(
            series[hours == hour_level[i]], [25.0, 75]
        )

    HourlyMedianResult = namedtuple(
        "HourlyMedianResult", ("hour_level", "median", "q1", "q3")
    )
    return HourlyMedianResult(hour_level, med_hr, q1_hr, q3_hr)


def hourly_avg(hours, series, all_hours=True, ddof=1):
    """
    Calculate hourly binned averages of a time series.

    Parameters
    ----------
    hours : array_like
        Time series of the hour number. Must be of the same length as `series`.
    series : array_like
        Time series of the data.
    all_hours : bool, optional
        Default is `True` to consider 24 hours. If `False`, only consider the
        hours that are present in the `hours` input.
    ddof : int, optional
        Degree of freedom for standard deviation calculation. Default is 1.

    Returns
    -------
    hour_level : array_like
        Hour levels.
    median : array_like
        Average values by the hour.
    std : array_like
        Standard deviation values by the hour.

    See Also
    --------
    `hourly_median` : Hourly binned median function.

    """
    if all_hours:
        hour_level = _np.arange(24)
    else:
        hour_level = _np.unique(hours)
    avg_hr = _np.zeros(hour_level.size) + _np.nan
    std_hr = _np.zeros(hour_level.size) + _np.nan
    for i in range(hour_level.size):
        avg_hr[i] = _np.nanmean(series[hours == hour_level[i]])
        std_hr[i] = _np.nanstd(series[hours == hour_level[i]], ddof=ddof)

    HourlyAverageResult = namedtuple(
        "HourlyAverageResult", ("hour_level", "avg", "std")
    )
    return HourlyAverageResult(hour_level, avg_hr, std_hr)


def simple_gapfill(x, y, left=None, right=None):
    """
    A simple linear gap-fill function.

    Parameters
    ----------
    x : array_like
        The time variable of the time series.
    y : array_like
        The time series to be gap-filled.
    left : float
        Left boundary value.
    right : float
        Right boundary value.

    Returns
    -------
    array_like
        The linearly gap-filled time series. Has the same length as the input
        time series.

    """
    xfinite = x[_np.isfinite(y)]
    yfinite = y[_np.isfinite(y)]
    return _np.interp(x, xfinite, yfinite, left=left, right=right)



"""Solar radiation."""
import datetime
from collections import namedtuple

import numpy as _np
import scipy.constants as _sc

from ecoflux.constants import constants


def solar_angle(dt, lat, lon, timezone=0.0):
    """
    Calculate the solar angle and sunrise/sunset times.

    Original program by `NOAA Global Radiation Group
    <http://www.esrl.noaa.gov/gmd/grad/solcalc/calcdetails.html>`_.

    Translated to Python by Wu Sun <wu.sun@ucla.edu> on 23 Nov 2015.

    Parameters
    ----------
    dt : datetime.datetime
        Time variable packaged in the built-in datetime format.
    lat : float
        Latitude (-90 to 90).
    lon : float
        Longitude (-180 to 180).
    timezone : float, optional
        Time zone with respect to UTC (-12 to 12), by default set to 0.

    Returns
    -------
    solar_angle_result : namedtuple
        Unpack the namedtuple fields to get results.
        - 'solar noon': local solar noon, in fraction of a day
        - 'sunrise': sunrise time, in fraction of a day
        - 'sunset': sunset time, in fraction of a day
        - 'hour angle': hour angle, in degree
        - 'solar zenith angle': solar zenith angle, in degree
        - 'solar elevation angle': solar elevation angle, in degree
        - 'solar azimuth angle': solar azimuth angle, in degree
        - 'atmospheric refraction': atmospheric refraction, in degree
        - 'solar zenith angle corrected': solar zenith angle corrected for
          atmospheric refraction, in degree
        - 'solar elevation angle corrected': solar elevation angle corrected
          for atmospheric refraction, in degree

    Raises
    ------
    TypeError
        If `dt` is not a `datetime.datetime` instance.

    """
    if not isinstance(dt, datetime.datetime):
        raise TypeError("Input `dt` is not a `datetime.datetime` instance.")

    # current eccentricity of earth orbit
    eccentricity = constants.eccentricity

    # use NASA truncated Julian Date method to calculate Julian Date number
    # JD = 2440000.5 at 00:00 on May 24, 1968
    time_delta = dt - datetime.datetime(1968, 5, 24, 0, 0, 0)
    julian_date = (
        2440000.5
        + time_delta.days
        + time_delta.seconds / 86400.0
        - timezone / 24.0
    )
    julian_century = (julian_date - 2451545.0) / 36525.0

    # geometric mean longitude of the sun, in degree
    geom_mean_lon_sun = _np.mod(
        280.46646
        + julian_century * (36000.76983 + julian_century * 0.0003032),
        360,
    )
    # geometric mean anomaly of the sun, in degree
    geom_mean_anom_sun = 357.52911 + julian_century * (
        35999.05029 - 0.0001537 * julian_century
    )
    # solar equator of center
    sun_eq_ctr = (
        _np.sin(_np.radians(geom_mean_anom_sun))
        * (1.914602 - julian_century * (0.004817 + 1.4e-5 * julian_century))
        + _np.sin(_np.radians(geom_mean_anom_sun) * 2)
        * (0.019993 - 0.000101 * julian_century)
        + _np.sin(_np.radians(geom_mean_anom_sun) * 3) * 0.000289
    )
    # solar true longitude
    sun_true_lon = geom_mean_lon_sun + sun_eq_ctr
    # # solar true anomaly
    # sun_true_anom = geom_mean_anom_sun + sun_eq_ctr
    # # solar rad vector in AUs
    # sun_rad_vector = (1.000001018 * (1. - eccentricity**2)) / \
    #     (1. + eccentricity * _np.cos(_np.radians(sun_true_anom)))
    # solar apparent longitude
    sun_app_lon = (
        sun_true_lon
        - 0.00569
        - 0.00478 * _np.sin(_np.radians(125.04 - 1934.136 * julian_century))
    )

    # in degree
    mean_obliq_ecliptic = (
        23.0
        + (
            26.0
            + (
                21.448
                - julian_century
                * (
                    46.815
                    + julian_century * (0.00059 - julian_century * 0.001813)
                )
            )
            / 60.0
        )
        / 60.0
    )

    obliq_corr = mean_obliq_ecliptic + 0.00256 * _np.cos(
        _np.radians(125.04 - 1934.136 * julian_century)
    )
    # sun_rt_ascen = _np.degrees(_np.arctan2(
    #     _np.cos(_np.radians(obliq_corr)) * _np.sin(_np.radians(sun_app_lon)),
    #     _np.cos(_np.radians(sun_app_lon))))
    sun_declin = _np.degrees(
        _np.arcsin(
            _np.sin(_np.radians(obliq_corr))
            * _np.sin(_np.radians(sun_app_lon))
        )
    )

    var_y = _np.tan(_np.radians(obliq_corr / 2.0)) ** 2
    eq_time = 4.0 * _np.degrees(
        var_y * _np.sin(2.0 * _np.radians(geom_mean_lon_sun))
        - 2.0 * eccentricity * _np.sin(_np.radians(geom_mean_anom_sun))
        + 4.0
        * eccentricity
        * var_y
        * _np.sin(_np.radians(geom_mean_anom_sun))
        * _np.cos(2.0 * _np.radians(geom_mean_lon_sun))
        - 0.5 * var_y ** 2 * _np.sin(4.0 * _np.radians(geom_mean_lon_sun))
        - 1.25
        * eccentricity ** 2
        * _np.sin(2.0 * _np.radians(geom_mean_anom_sun))
    )
    # in minutes

    HA_sunrise = _np.degrees(
        _np.arccos(
            _np.cos(_np.radians(90.833))
            / (_np.cos(_np.radians(lat)) * _np.cos(_np.radians(sun_declin)))
            - _np.tan(_np.radians(lat)) * _np.tan(_np.radians(sun_declin))
        )
    )

    solar_noon_local = (
        720.0 - 4.0 * lon - eq_time + timezone * 60.0
    ) / 1440.0  # in day
    sunrise_local = solar_noon_local - HA_sunrise * 4.0 / 1440.0  # in day
    sunset_local = solar_noon_local + HA_sunrise * 4.0 / 1440.0  # in day
    # sunlight_duration = 8. * HA_sunrise / 1440.  # in day
    # sunlight_duration_min = 8. * HA_sunrise  # in minutes

    fractional_day = (
        dt.hour / 24.0
        + dt.minute / 1440.0
        + (dt.second + dt.microsecond * 1e-6) / 86400.0
    )
    # true solar time in minutes
    true_solar_time_min = _np.mod(
        fractional_day * 1440.0 + eq_time + 4.0 * lon - 60.0 * timezone, 1440
    )

    # calculate hour angle
    if true_solar_time_min / 4.0 < 0.0:
        hour_angle = true_solar_time_min / 4.0 + 180.0
    else:
        hour_angle = true_solar_time_min / 4.0 - 180.0

    solar_zenith_angle = _np.degrees(
        _np.arccos(
            _np.sin(_np.radians(lat)) * _np.sin(_np.radians(sun_declin))
            + _np.cos(_np.radians(lat))
            * _np.cos(_np.radians(sun_declin))
            * _np.cos(_np.radians(hour_angle))
        )
    )

    solar_elev_angle = 90.0 - solar_zenith_angle

    # calculate solar azimuth angle
    if hour_angle > 0.0:
        solar_azimuth_angle = _np.mod(
            _np.degrees(
                _np.arccos(
                    (
                        (
                            _np.sin(_np.radians(lat))
                            * _np.cos(_np.radians(solar_zenith_angle))
                        )
                        - _np.sin(_np.radians(sun_declin))
                    )
                    / (
                        _np.cos(_np.radians(lat))
                        * _np.sin(_np.radians(solar_zenith_angle))
                    )
                )
            )
            + 180,
            360,
        )
    else:
        solar_azimuth_angle = _np.mod(
            540.0
            - _np.degrees(
                _np.arccos(
                    (
                        (
                            _np.sin(_np.radians(lat))
                            * _np.cos(_np.radians(solar_zenith_angle))
                        )
                        - _np.sin(_np.radians(sun_declin))
                    )
                    / (
                        _np.cos(_np.radians(lat))
                        * _np.sin(_np.radians(solar_zenith_angle))
                    )
                )
            ),
            360,
        )

    if solar_elev_angle > 85.0:
        approx_atmos_refrac = 0.0
    elif 5.0 < solar_elev_angle <= 85.0:
        approx_atmos_refrac = (
            58.1 / _np.tan(_np.radians(solar_elev_angle))
            - 0.07 / _np.tan(_np.radians(solar_elev_angle)) ** 3
            + 0.000086 / _np.tan(_np.radians(solar_elev_angle)) ** 5
        ) / 3600.0
    elif -0.575 < solar_elev_angle <= 5.0:
        approx_atmos_refrac = (
            1735.0
            - 518.2 * solar_elev_angle
            + 103.4 * solar_elev_angle ** 2
            - 12.79 * solar_elev_angle ** 3
            + 0.711 * solar_elev_angle ** 4
        ) / 3600.0
    elif solar_elev_angle <= -0.575:
        approx_atmos_refrac = (
            -20.774 / 3600.0 / _np.tan(_np.radians(solar_elev_angle))
        )

    solar_zenith_angle_corr = solar_zenith_angle - approx_atmos_refrac
    solar_elev_angle_corr = solar_elev_angle + approx_atmos_refrac

    SolarAngleResult = namedtuple(
        "SolarAngleResult",
        (
            "solar_noon",
            "sunrise",
            "sunset",
            "hour_angle",
            "solar_zenith_angle",
            "solar_elevation_angle",
            "solar_azimuth_angle",
            "atmospheric_refraction",
            "solar_zenith_angle_corrected",
            "solar_elevation_angle_corrected",
        ),
    )
    solar_angle_result = SolarAngleResult(
        solar_noon_local,
        sunrise_local,
        sunset_local,
        hour_angle,
        solar_zenith_angle,
        solar_elev_angle,
        solar_azimuth_angle,
        approx_atmos_refrac,
        solar_zenith_angle_corr,
        solar_elev_angle_corr,
    )

    return solar_angle_result


def planck_law(wavelength, temp, emissivity=1.0, kelvin=False):
    """
    Calculate spectral radiance from Planck's law.

    Parameters
    ----------
    wavelength : float or array_like
        Wavelength of the photon [m]
    temp : float or array_like
        Temperature of the radiating body. Default unit is Celsius. To use
        Kelvin scale, set `kelvin=True`.
    emissivity : float or array_like, optional
        Emissivity of the radiating body [0 to 1]. Default is 1 for blackbody.
    kelvin : boolean, optional
        If False (default), temperature argument is treated as in Celsius;
        if True, temperature argument is treated as in Kelvin.

    Returns
    -------
    B_lambda: float or array_like
        Spectral radiance [W sr^-1 m^-3]

    """
    T_k = temp + (not kelvin) * _sc.zero_Celsius
    c1 = 2.0 * _sc.h * _sc.c ** 2
    c2 = _sc.h * _sc.c / _sc.k
    try:
        B_lambda = (
            emissivity
            * c1
            * wavelength ** (-5.0)
            / (_np.exp(c2 / wavelength / T_k) - 1.0)
        )
    except ZeroDivisionError:
        # spectral radiance is zero at 0 K
        # but needs to keep the array size the same
        B_lambda = 0.0 * (wavelength + T_k)
    return B_lambda



"""Canopy radiative transfer."""
import numpy as _np


def extinc_coef(theta, chi):
    """
    Calculate the extinction coefficient of the plant canopy.

    Parameters
    ----------
    theta : float or array_like
        Solar zenith angle in radians. Must be within [0, pi / 2).
    chi : float or array_like
        Leaf shape parameter for light extinction.

    Returns
    -------
    float or array_like
        The light extinction coefficient of the canopy [m^2 m^-2].

    """
    return _np.sqrt(chi * chi + _np.tan(theta) ** 2) / (
        chi + 1.774 * (chi + 1.182) ** (-0.733)
    )



"""A collection of functions of statistical distributions."""
import numpy as _np


def binsize(x):
    """
    Calculate the optimal bin size for histograms following the
    Freedman-Diaconis rule [FD81]_.

    Parameters
    ----------
    x : array_like
        The input data. Must be one-dimensional.

    Returns
    -------
    float
        The recommended bin size for the input data.

    References
    ----------
    .. [FD81] Freedman, D. and Diaconis, P. (1981). On the histogram as a
    density estimator: L2 theory. *Zeitschrift für Wahrscheinlichkeitstheorie
    und verwandte Gebiete*, 57(4), 453-476.

    """
    xfinite = x[_np.isfinite(x)]
    q1, q3 = _np.percentile(xfinite, [25.0, 75.0])
    return (q3 - q1) * 2.0 / xfinite.size ** (1.0 / 3.0)



"""
============================================
Radiative transfer (:mod:`ecoflux.radtrans`)
============================================

.. currentmodule:: ecoflux.radtrans

Functions to calculate the radiative transfer in the atmosphere and canopies.

"""
from . import canopy_light, solar_radiation  # noqa



"""Light response of leaf photosynthesis."""
from typing import Tuple

import numpy as _np


def michaelis_menten(p: Tuple[float, float, float], par: _np.ndarray):
    r"""
    Michaelis–Menten light response function.

    Parameters
    ----------
    p : tuple
        A tuple of three parameters

        * p[0]: *K*\ :sub:`PAR`, Michaelis constant
          [µmol photons m\ :sup:`–2` s\ :sup:`–1`];
        * p[1]: *P*\ :sub:`m`, maximum gross photosynthetic rate
          [µmol m\ :sup:`–2` s\ :sup:`–1`];
        * p[2]: *R*\ :sub:`d`, daytime respiration rate
          [µmol m\ :sup:`–2` s\ :sup:`–1`].

    par : array_like
        Photosynthetically active radiation
        [µmol photons m\ :sup:`–2` s\ :sup:`–1`].

    Returns
    -------
    array_like
        Photosynthetic assimilation rate [µmol m\ :sup:`–2` s\ :sup:`–1`].

    Examples
    --------
    >>> import numpy as np
    >>> michaelis_menten([500., 20., 2.], np.array([1500., 2000.]))
    array([13., 14.])

    """
    k_par, p_m, r_d = p
    return p_m * par / (k_par + par) - r_d


def residual_michaelis_menten(
    p: Tuple[float, float, float], par: _np.ndarray, A_n: _np.ndarray,
):
    r"""
    Residual function for Michaelis–Menten light response.

    Parameters
    ----------
    p : tuple
        A tuple of three parameters

        * p[0]: *K*\ :sub:`PAR`, Michaelis constant
          [µmol photons m\ :sup:`–2` s\ :sup:`–1`];
        * p[1]: *P*\ :sub:`m`, maximum gross photosynthetic rate
          [µmol m\ :sup:`–2` s\ :sup:`–1`];
        * p[2]: *R*\ :sub:`d`, daytime respiration rate
          [µmol m\ :sup:`–2` s\ :sup:`–1`].

    par : array_like
        Photosynthetically active radiation
        [µmol photons m\ :sup:`–2` s\ :sup:`–1`].
    A_n : array_like
        Measured photosynthetic assimilation rate
        [µmol m\ :sup:`–2` s\ :sup:`–1`].

    Returns
    -------
    array_like
        Residual of the photosynthetic assimilation rate
        [µmol m\ :sup:`–2` s\ :sup:`–1`].

    """
    k_par, p_m, r_d = p
    return p_m * par / (k_par + par) - r_d - A_n


def hyperbolic(p: Tuple[float, float, float, float], par: _np.ndarray):
    r"""
    Hyperbolic light response function.

    Parameters
    ----------
    p : tuple
        A tuple of four parameters

        * p[0]: *θ*, a curvature parameter;
        * p[1]: *α*, apparent quantum yield;
        * p[2]: *P*\ :sub:`m`, maximum gross photosynthetic rate
          [µmol m\ :sup:`–2` s\ :sup:`–1`];
        * p[3]: *R*\ :sub:`d`, daytime respiration rate
          [µmol m\ :sup:`–2` s\ :sup:`–1`].

    par : array_like
        Photosynthetically active radiation
        [µmol photons m\ :sup:`–2` s\ :sup:`–1`].

    Returns
    -------
    A_n : array_like
        Photosynthetic assimilation rate [µmol m\ :sup:`–2` s\ :sup:`–1`].

    Examples
    --------
    >>> import numpy as np
    >>> hyperbolic([0.7, 0.04, 20., 2.], np.array([1500., 2000.]))
    array([15.75986071, 16.35949823])

    References
    ----------
    * Ögren, E. and Evans, J. R. (1993). Photosynthetic light-response curves:
      I. The influence of CO2 partial pressure and leaf inversion. *Planta*,
      189, 182–190. https://doi.org/10.1007/BF00195075

    """
    theta, alpha, p_m, r_d = p
    if _np.isclose(theta, 0.0):
        A_n = alpha * p_m * par / (alpha * par + p_m) - r_d
    else:
        A_n = (
            alpha * par
            + p_m
            - _np.sqrt(
                (alpha * par + p_m) ** 2 - 4.0 * theta * alpha * par * p_m
            )
        ) * 0.5 / theta - r_d
    return A_n


def residual_hyperbolic(
    p: Tuple[float, float, float, float], par: _np.ndarray, A_n: _np.ndarray,
):
    r"""
    Residual function for the hyperbolic light response.

    Parameters
    ----------
    p : tuple
        A tuple of four parameters

        * p[0]: *θ*, a curvature parameter;
        * p[1]: *α*, apparent quantum yield;
        * p[2]: *P*\ :sub:`m`, maximum gross photosynthetic rate
          [µmol m\ :sup:`–2` s\ :sup:`–1`];
        * p[3]: *R*\ :sub:`d`, daytime respiration rate
          [µmol m\ :sup:`–2` s\ :sup:`–1`].

    par : array_like
        Photosynthetically active radiation
        [µmol photons m\ :sup:`–2` s\ :sup:`–1`].
    A_n : array_like
        Measured photosynthetic assimilation rate
        [µmol m\ :sup:`–2` s\ :sup:`–1`].

    Returns
    -------
    array_like
        Residual of the photosynthetic assimilation rate
        [µmol m\ :sup:`–2` s\ :sup:`–1`].

    """
    return hyperbolic(p, par) - A_n



"""
====================================
Constants (:mod:`ecoflux.constants`)
====================================

.. currentmodule:: ecoflux.constants

Physical, chemical, and environmental constants.

Gases
=====

============  ==========================================================
``M_d``       dry air molar mass [kg mol^-1]
``M_w``       water vapor molar mass [kg mol^-1]
``R_d``       specific gas constant of dry air [J kg^-1 K^-1]
``R_w``       specific gas constant of water vapor[J kg^-1 K^-1]
``cp_d``      isobaric specific heat capacity of dry air [J kg^-1 K^-1]
``cpm_d``     isobaric molar heat capacity of dry air [J mol^-1 K^-1]
============  ==========================================================

Soil
====

===================  =======================================================
``soil_textures``    soil texture names according to the USDA classification
===================  =======================================================

Earth
=====

==================  =============================================
``eccentricity``    present-day eccentricity of the earth's orbit
==================  =============================================

Atmosphere
==========

===============  =================================
``m_atm``        total mass of the atmosphere [kg]
``m_atm_d``      dry mass of the atmosphere [kg]
``Gamma_d``      dry lapse rate [K m^-1]
``Gamma_mean``   mean lapse rate [K m^-1]
``kappa``        von Karman constant [-]
===============  =================================

References
==========

* Or, D. and Wraith, J. M. (2002). Soil Water Content and Water Potential
  Relationships, in Warrick, A. W. (eds.) *Soil Physics Companion*, pp 81–82.,
  CRC Press, Boca Raton, FL, USA.
* Rumble, J. (eds.) (2017). *CRC Handbook of Chemistry and Physics* (98th ed.).
  CRC Press, Boca Raton, FL, USA. ISBN: 9781498784542.
  http://hbcponline.com/faces/contents/ContentsSearch.xhtml
* Trenberth, K. E. and Smith, L. (2005). The mass of the atmosphere: a
  constraint on global analyses. *Journal of Climate*, 18, 864–875.
  https://doi.org/10.1175/JCLI-3299.1

"""
from .constants import *  # noqa

__all__ = [_s for _s in dir() if not _s.startswith("_")]



"""A collection of functions for summary statistics."""
import numpy as _np


def mad(x):
    """
    Calculate the median absolute deviation.

    Parameters
    ----------
    x : array_like
        The sample

    Returns
    -------
    float
        The median absolute deviation of the sample.

    """
    return _np.nanmedian(_np.abs(x - _np.nanmedian(x)))


def zscore(x, robust_zscore=False):
    """
    Calculate Z-score of the sample.

    Parameters
    ----------
    x : array_like
        The sample
    robust_zscore : bool, optional
        If `True`, use the Iglewicz-Hoaglin robust Z-score [IH93]_ instead of
        the original Z-score. Default is `False`.

    Returns
    -------
    float
        The calculated Z-score of the sample.

    References
    ----------
    .. [IH93] Iglewicz, Boris and Hoaglin, David C. (1993). How to Detect and
        Handle Outliers. ASQC Quality Press, Milwaukee, WI, 1993.

    """
    if not robust_zscore:
        return (x - _np.nanmean(x)) / _np.nanstd(x, ddof=1)
    else:
        return 0.6745 * (x - _np.nanmedian(x)) / mad(x)


def interquartile(x, axis=None):
    """
    Calculate the interquartile range of an array.

    Parameters
    ----------
    x : array_like
        The sample.
    axis : int, optional
        Axis along which the percentiles are computed. Default is to ignore
        and compute the flattened array.
        (Same as the `axis` argument in `numpy.nanpercentile()`.)

    Returns
    -------
    float or array_like
        The interquartile range of the sample.

    """
    if _np.sum(_np.isfinite(x)) > 0:
        q1, q3 = _np.nanpercentile(x, [25.0, 75.0], axis=axis)
        return q3 - q1
    else:
        return _np.nan


def resist_mean(x, inlier_range=1.5):
    """
    Calculate outlier-resistant mean of the sample using Tukey's outlier test.

    Warning: Does not support calculation along an axis, unlike `numpy.mean()`.

    Parameters
    ----------
    x : array_like
        The sample. Must be one-dimensional.
    inlier_range : float, optional
        Parameter to control the inlier range defined by
        [Q_1 - inlier_range * (Q_3 - Q_1), Q_3 - inlier_range * (Q_3 - Q_1)]
        Default value is 1.5.

    Returns
    -------
    rmean : float
        The resistant mean of the sample with outliers removed.

    References
    ----------
    .. [T77] John W. Tukey (1977). Exploratory Data Analysis. Addison-Wesley.

    """
    x = _np.array(x)
    if _np.sum(_np.isfinite(x)) <= 1:
        return _np.nanmean(x)
    else:
        q1, q3 = _np.nanpercentile(x, [25.0, 75.0])
        iqr = q3 - q1
        uplim = q3 + inlier_range * iqr
        lolim = q1 - inlier_range * iqr
        rmean = _np.nanmean(x[(x >= lolim) & (x <= uplim)])
        return rmean


def resist_std(x, inlier_range=1.5):
    """
    Calculate outlier-resistant standard deviation of the sample using
    Tukey's outlier test.

    Warning: Does not support calculation along an axis, unlike `numpy.std()`.

    Parameters
    ----------
    x : array_like
        The sample. Must be one-dimensional.
    inlier_range : float, optional
        Parameter to control the inlier range defined by
        [Q_1 - inlier_range * (Q_3 - Q_1), Q_3 - inlier_range * (Q_3 - Q_1)]
        Default value is 1.5.

    Returns
    -------
    rstd : float
        The resistant standard deviation of the sample with outliers removed.
        Degree of freedom = 1 is enforced for the sample standard deviation.

    References
    ----------
    .. [T77] John W. Tukey (1977). Exploratory Data Analysis. Addison-Wesley.

    """
    x = _np.array(x)
    if _np.sum(_np.isfinite(x)) <= 1:
        return _np.nanstd(x, ddof=1)
    else:
        q1, q3 = _np.nanpercentile(x, [25.0, 75.0])
        iqr = q3 - q1
        uplim = q3 + inlier_range * iqr
        lolim = q1 - inlier_range * iqr
        rstd = _np.nanstd(x[(x >= lolim) & (x <= uplim)], ddof=1)
        return rstd


def dixon_test(x, left=True, right=True, q_conf="q95"):
    """
    Use Dixon's Q test to identify one or two outliers. The test is based upon
    two assumptions: (1) data must be normally distributed; (2) the test may
    only be used once to a dataset and not repeated.

    Adapted from: <http://sebastianraschka.com/Articles/2014_dixon_test.html>
    (Retrieved 23 Apr 2017).

    Parameters
    ----------
    x : array_like
        Data points. Must be one-dimensional.
    left : bool, optional
        If True, test the minimum value.
    right : bool, optional
        If True, test the maximum value.
        (At least one of the two, `left` or `right`, must be True.)
    q_conf : str, optional
        Confidence level: 'q95' -- 95% confidence (default). Other options
        supported are 'q90' (90% C.I.) and 'q99' (99% C.I.).

    Returns
    -------
    outliers : list
        A list of two values containing the outliers or None. The first element
        corresponds to the minimum value, and the second element corresponds
        to the maximum value. If the tested value is not an outlier, return
        None at its position.

    References
    ----------
    .. [1] Dean, R. B. and Dixon, W. J. (1951). Simplified Statistics for Small
       Numbers of Observations. Anal. Chem., 23(4), 636—638.
    .. [2] Dixon, W. J. (1953). Processing data for outliers Reference.
       J. Biometrics, 9, 74–89.
    .. [3] Rorabacher, D. B. (1991). Statistical Treatment for Rejection of
       Deviant Values: Critical Values of Dixon Q Parameter and Related
       Subrange Ratios at the 95 percent Confidence Level. Anal. Chem., 63(2),
       139–146.

    """
    # critical Q value table
    q_dicts = {
        "q90": [
            0.941,
            0.765,
            0.642,
            0.560,
            0.507,
            0.468,
            0.437,
            0.412,
            0.392,
            0.376,
            0.361,
            0.349,
            0.338,
            0.329,
            0.320,
            0.313,
            0.306,
            0.300,
            0.295,
            0.290,
            0.285,
            0.281,
            0.277,
            0.273,
            0.269,
            0.266,
            0.263,
            0.260,
        ],
        "q95": [
            0.970,
            0.829,
            0.710,
            0.625,
            0.568,
            0.526,
            0.493,
            0.466,
            0.444,
            0.426,
            0.410,
            0.396,
            0.384,
            0.374,
            0.365,
            0.356,
            0.349,
            0.342,
            0.337,
            0.331,
            0.326,
            0.321,
            0.317,
            0.312,
            0.308,
            0.305,
            0.301,
            0.290,
        ],
        "q99": [
            0.994,
            0.926,
            0.821,
            0.740,
            0.680,
            0.634,
            0.598,
            0.568,
            0.542,
            0.522,
            0.503,
            0.488,
            0.475,
            0.463,
            0.452,
            0.442,
            0.433,
            0.425,
            0.418,
            0.411,
            0.404,
            0.399,
            0.393,
            0.388,
            0.384,
            0.38,
            0.376,
            0.372,
        ],
    }
    # cast to numpy array and remove NaNs
    x_arr = _np.array(x)
    x_arr = x_arr[_np.isfinite(x_arr)]
    # minimum and maximum data sizes allowed
    min_size = 3
    max_size = len(q_dicts[q_conf]) + min_size - 1
    if len(x_arr) < min_size:
        raise ValueError(
            "Sample size too small: "
            + "at least %d data points are required" % min_size
        )
    elif len(x_arr) > max_size:
        raise ValueError("Sample size too large")

    if not (left or right):
        raise ValueError(
            "At least one of the two options, "
            + "`left` or `right`, must be True."
        )

    q_crit = q_dicts[q_conf][len(x_arr) - 3]

    # for small dataset, the built-in `sorted()` is faster than `np.sort()`
    x_sorted = sorted(x_arr)

    x_range = x_sorted[-1] - x_sorted[0]
    if x_range == 0:
        outliers = [None, None]
    else:
        Q_min = abs((x_sorted[1] - x_sorted[0]) / x_range)
        Q_max = abs((x_sorted[-1] - x_sorted[-2]) / x_range)
        outliers = [
            x_sorted[0] if (Q_min > q_crit) and (Q_min >= Q_max) else None,
            x_sorted[-1] if (Q_max > q_crit) and (Q_max >= Q_min) else None,
        ]

    return outliers



"""A collection of physical, chemical, and environmental constants."""
from typing import List

import scipy.constants as _sc

# chemical constants
M_d: float = 28.964_5e-3  # dry air molar mass [kg mol^-1]
M_w: float = 18.015_28e-3  # water vapor molar mass [kg mol^-1]
R_d: float = _sc.R / M_d  # specific gas constant of dry air [J kg^-1 K^-1]
R_w: float = _sc.R / M_w  # specific gas constant of water vapor[J kg^-1 K^-1]

# isobaric specific heat capacity of dry air [J kg^-1 K^-1]
cp_d: float = 1.004e3
# isobaric molar heat capacity of dry air [J mol^-1 K^-1]
cpm_d: float = cp_d * M_d
# note: variation with temperature is negligible in the atmosphere

# soil texture names, USDA classification
soil_textures: List[str] = [
    "sand",
    "loamy sand",
    "sandy loam",
    "loam",
    "silt",
    "silt loam",
    "sandy clay loam",
    "clay loam",
    "silty clay loam",
    "sandy clay",
    "silty clay",
    "clay",
]

# properties of the earth
eccentricity: float = 0.016_704_232

# properties of the atmosphere
m_atm: float = 5.1480e18  # total mass of the atmosphere [kg]
m_atm_d: float = 5.1352e18  # dry mass of the atmosphere [kg]
Gamma_d: float = _sc.g / cp_d  # dry lapse rate [K m^-1]
Gamma_mean: float = 6.5e-3  # mean lapse rate [K m^-1]
kappa: float = 0.40  # von Karman constant [-]



"""Saturation vapor pressure of water."""
import numpy as _np
import scipy.constants as _sc
from scipy import optimize as _optimize

T_0: float = _sc.zero_Celsius


def p_sat_h2o(temp, ice=False, kelvin=False, method="gg"):
    """
    Calculate saturation vapor pressure over water or ice at a temperature.

    by Wu Sun <wu.sun@ucla.edu>, 14 Sep 2014

    Parameters
    ----------
    temp : float or array_like
        Temperature, in Celsius degree by default.
    ice : bool, optional
        Calculate saturation vapor pressure on ice if enabled.
    kelvin : bool, optional
        Temperature input is in Kelvin if enabled.
    method : str, optional
        Method used to evaluate saturation vapor pressure.
        'gg': default, Goff-Gratch equation (1946). [GG46]_
        'buck': Buck Research Instruments L.L.C. (1996). [B96]_
        'cimo': CIMO Guide (2008). [WMO]_

    Returns
    -------
    e_sat : float or array_like
        Saturation vapor pressure in Pascal.

    Raises
    ------
    ValueError
        If keyword 'ice' is enabled but temperature is above 0 C or 273.15 K.

    References
    ----------
    .. [GG46] Goff, J. A., and Gratch, S. (1946). Low-pressure properties of
       water from -160 to 212 F, in Transactions of the American Society of
       Heating and Ventilating Engineers, pp 95-122, presented at the 52nd
       Annual Meeting of the American Society of Heating and Ventilating
       Engineers, New York.
    .. [B96] Buck Research Instruments L.L.C. (1996). *Buck Research CR-1A
       User's Manual*, Appendix 1.
    .. [WMO] World Meteorological Organization. (2008). *Guide to
       Meteorological Instruments and Methods of Observation*, Appendix 4B,
       WMO-No. 8 (CIMO Guide), Geneva.

    Examples
    --------
    >>> p_sat_h2o(25)
    3165.195633383682

    >>> p_sat_h2o([0, 5, 15, 25])
    array([  610.33609993,   871.31372986,  1703.28100711,  3165.19563338])

    >>> p_sat_h2o(25, method='buck')
    3168.5314122754344

    >>> p_sat_h2o(273.15, kelvin=True)
    610.33609993341383

    >>> p_sat_h2o(-15, ice=True)
    165.01477392358936

    >>> p_sat_h2o(258.15, kelvin=True, ice=True, method='cimo')
    165.28713201714956

    """
    T_k = _np.array(temp, dtype="d") + (not kelvin) * T_0
    # force temperature to be in Kelvin

    if _np.sum(T_k > 273.16) and ice:
        # The triple point of water is 273.16 K
        raise ValueError("Temperature error, no ice exists.")

    if not ice:
        if method == "buck":
            T_c = T_k - T_0  # temperature in Celsius degree
            e_sat = (
                6.1121
                * _np.exp((18.678 - T_c / 234.5) * T_c / (257.14 + T_c))
                * 100
            )
        elif method == "cimo":
            T_c = T_k - T_0  # temperature in Celsius degree
            e_sat = 6.112 * _np.exp(17.62 * T_c / (243.12 + T_c)) * 100
        else:
            # Goff-Gratch equation by default
            u_T = 373.16 / T_k
            v_T = T_k / 373.16
            e_sat = (
                -7.90298 * (u_T - 1)
                + 5.02808 * _np.log10(u_T)
                - 1.3816e-7 * (10 ** (11.344 * (1 - v_T)) - 1)
                + 8.1328e-3 * (10 ** (-3.49149 * (u_T - 1)) - 1)
                + _np.log10(1013.246)
            )
            e_sat = 10 ** e_sat * 100
    else:
        if method == "buck":
            T_c = T_k - T_0  # temperature in Celsius degree
            e_sat = (
                6.1115
                * _np.exp((23.036 - T_c / 333.7) * T_c / (279.82 + T_c))
                * 100
            )
        elif method == "cimo":
            T_c = T_k - T_0  # temperature in Celsius degree
            e_sat = 6.112 * _np.exp(22.46 * T_c / (272.62 + T_c)) * 100
        else:
            # Goff-Gratch equation by default
            u_T = 273.16 / T_k
            v_T = T_k / 273.16
            e_sat = (
                -9.09718 * (u_T - 1)
                - 3.56654 * _np.log10(u_T)
                + 0.876793 * (1 - v_T)
                + _np.log10(6.1071)
            )
            e_sat = 10 ** e_sat * 100

    return e_sat


def dew_temp(e_sat, guess=25.0, kelvin=False, method="gg"):
    """
    Calculate dew temperature from water concentration.

    Parameters
    ----------
    e_sat : float
        Saturation vapor pressure in Pascal. Takes only a single value,
        no array allowed.
    guess : float, optional
        An initial guess for the dew temperature to infer.
    kelvin : bool, optional
        Dew temperature calculated is in Kelvin if enabled.
    method : str, optional
        Method used to evaluate saturation vapor pressure.
        'gg': default, Goff-Gratch equation (1946). [GG46]_
        'buck': Buck Research Instruments L.L.C. (1996). [B96]_
        'cimo': CIMO Guide (2008). [WMO]_

    Returns
    -------
    T_dew : float
        Dew temperature.

    Examples
    --------
    >>> dew_temp(3165)
    24.998963153421172

    >>> dew_temp(610)
    -0.0075798295337097081

    >>> dew_temp(3165, kelvin=True)
    298.14896315342116

    """

    def __e_sat_residual(T, e_sat, kelvin, method):
        return p_sat_h2o(T, kelvin=kelvin, method=method) - e_sat

    if kelvin:
        guess += T_0

    try:
        T_dew = _optimize.newton(
            __e_sat_residual, x0=guess, args=(e_sat, kelvin, method)
        )
    except RuntimeError:
        T_dew = _np.nan

    return T_dew



"""
============================================
Statistical functions (:mod:`ecoflux.stats`)
============================================

.. currentmodule:: ecoflux.stats

Statistical functions used in processing flux data.

"""
from . import dists, regressions, summary, timeseries  # noqa
